@@ -10,7 +10,10 @@ use warnings;
 use Debian::Debhelper::Dh_Lib;
 use File::Find;
 use JSON;
+use Graph;
 use IPC::Run qw(run);
+use Debian::PkgJs::PackageLock;
+use Debian::PkgJs::Utils;
 use parent qw(Debian::Debhelper::Buildsystem);
 
 use constant MAX_BUILD_LOOP => 3;
@@ -19,6 +22,8 @@ sub NPMIGNOREDEFAULT {
     $ENV{NPMIGNOREDEFAULT} || '/usr/share/pkg-js-tools/npmignore.default';
 }
 
+set_debug(1);
+
 # echo, touch and ls are here to help tests
 my @known_build_commands = (qw(grunt echo touch ls));
 
@@ -30,14 +35,16 @@ my @knwonBuildFiles = (
     [ 'Gruntfile.coffee' => ['grunt'] ],
 
     #[ 'tsconfig.json'     => ['tsc'] ],
-    #[ 'gulpfile.js' => ['gulp'] ],
-    #[ 'rollup.config.js'  => [ 'rollup', '-c' ] ],
-    #[ 'webpack.config.js' => ['webpack'] ],
-    #[ 'webpackfile.js'    => ['webpack'] ],
+    [ 'gulpfile.js'       => ['gulp'] ],
+    [ 'rollup.config.js'  => [ 'rollup',  '-c' ] ],
+    [ 'webpack.config.js' => [ 'webpack', '--mode', 'production' ] ],
+    [ 'webpackfile.js'    => [ 'webpack', '--mode', 'production' ] ],
 );
 
 my $knownCommandMap = { babel => 'babeljs', };
 
+my $noexecFiles = qr/\.(?:[cm]?js|json|[cm]?ts)/;
+
 sub DEBUG     { $ENV{DEB_BUILD_PKG_JS_DEBUG} }
 sub DEBUGFILE { $ENV{DEB_BUILD_PKG_JS_DEBUG_FILE} }
 
@@ -49,100 +56,19 @@ sub DESCRIPTION {
 
 # Regexp compiled from dh_nodejs/dh_root_files.excluded
 my $rootFilesIgnored =
-qr/^(?:(?:g(?:ulpfile\.(?:babel\.j|(?:t|j))|runt(?:file)?\.j)|(?:rollup[\.\-].*config|karma\.conf|.mocharc)\.j)s|b(?:abel\.config\.js(?:on)?|inding\.gyp|ench.*\.js|ower\.json)|c(?:o(?:mponent\.json|ntribute|pying)|hange(?:log|s?)|akefile)|a(?:ut(?:otest\.watchr|hors?)|va\.config\.js|ppveyor\.yml)|j(?:est\.config\.js|sl\.node\.conf)|.*\.(?:m(?:arkdown|d)|pdf|txt)|t(?:sconfig\.json|ests?\.js)|l(?:icen[cs]e.*|erna\.json)|\.(?:babelrc(?:.js)?|.*)|(?:Docker|make)file|package-lock\.json|yarn\.lock|history)$/i;
+qr/^(?:(?:g(?:ulpfile\.(?:babel\.m?j|(?:t|m?j))|runt(?:file)?\.j)|(?:rollup[\.\-].*config|karma\.conf|.mocharc)\.j)s|b(?:abel\.config\.js(?:on)?|inding\.gyp|ench.*\.js|ower\.json)|c(?:o(?:mponent\.json|ntribute|pying)|hange(?:log|s?)|akefile)|a(?:ut(?:otest\.watchr|hors?)|va\.config\.js|ppveyor\.yml)|j(?:est\.config\.js|sl\.node\.conf|asmine\.json)|.*\.(?:m(?:arkdown|d)|pdf|txt)|t(?:sconfig.*\.json|ests?\.js)|l(?:icen[cs]e.*|erna\.json)|\.(?:babelrc(?:.js)?|.*)|(?:Docker|make)file|package-lock\.json|yarn\.lock|history)$/i;
 my $filesIgnored =
 qr/^(?:\.(?:(?:wafpickle-|eslint|_).*|npm(?:ignore|rc)|lock-wscript|gitignore)|.*\.(?:c(?:offee|\+\+|cp?)?|orig|def|mk|h)|(?:[jm]|binding.M)akefile|readme(?:\.(?:txt|md))?|package-lock\.json|npm-debug\.log|yarn\.lock)$/i;
 
 # Regexp compiled from dh_nodejs/dh_dirs.excluded
 my $dirIgnored =
-qr{^(?:(.*/)?(?:\.(?:_.*|git|svn|hg)|.(?:DS_Store|deps)|__[^/]*__|fixtures?|CVS)|(?:archived-packag|node_modul)es|(?:example|test|doc)s?|bench|\..*)(?:/.*)?}i;
+qr{^(?:(?:.*/)?(?:\.(?:_.*|git|svn|hg)|.(?:DS_Store|deps)|__[^/]*__|fixtures?|CVS)|(?:archived-packag|node_modul)es|(?:example|test|doc)s?|bench|\..*)(?:/.*)?}i;
 
 # Extensions that overrides $rootFilesIgnored if set in package.json#files
 my $authorizedExt = qr/\.(?:js)$/;
 
 ### PRIVATE METHODS
 
-sub open_file {
-    my ( $self, $file ) = @_;
-    error unless $file;
-    my $f;
-    open( $f, $file ) or error $!;
-    my @lines = map {
-        chomp;
-        s/#.*$//;
-        s/^\s*(.*?)\s*/$1/;
-        $_ ? $_ : ();
-    } <$f>;
-    close $f;
-    return @lines;
-}
-
-my ( %json, $main_p );
-
-sub main_package {
-    my ($self) = @_;
-    return ( $main_p ? $main_p : () ) if defined $main_p;
-    return '.' unless -e 'debian/nodejs/main';
-    my @lines = $self->open_file('debian/nodejs/main');
-    error 'Malformed debian/nodejs/main'       if @lines > 1;
-    warning "Missing main directory $lines[0]" if @lines and !-d $lines[0];
-    $main_p = $lines[0] || '';
-    return @lines;
-}
-
-# package.json cache
-sub pjson {
-    my ( $self, $dir ) = @_;
-    unless ( -e "$dir/package.json" ) {
-        warning "/!\\ $dir/package.json not found\n";
-        return undef;
-    }
-    return $json{$dir} if $json{$dir};
-    my $pkgjson;
-    open $pkgjson, "$dir/package.json";
-    my $content = join '', <$pkgjson>;
-    close $pkgjson;
-    $json{$dir} = JSON::from_json($content);
-    if ( -e "debian/nodejs/$dir/name" ) {
-        $json{$dir}->{name} =
-          [ $self->open_file("debian/nodejs/$dir/name") ]->[0];
-    }
-    return $json{$dir};
-}
-
-sub component_list {
-    my $self = shift;
-    my ( @components, %packages );
-    if ( -e 'debian/watch' ) {
-        map { push @components, $1 if (/component=([\w\-\.]+)/) }
-          $self->open_file('debian/watch');
-    }
-    if ( -e 'debian/nodejs/additional_components' ) {
-        print "Found debian/nodejs/additional_components\n";
-        map {
-            print 'Adding component(s): ' . join( ', ', glob($_) ) . "\n";
-            push @components, glob($_)
-        } $self->open_file('debian/nodejs/additional_components');
-    }
-    foreach my $component (@components) {
-        next if $self->main_package and $component eq $self->main_package;
-        if ( -d $component ) {
-            my $package;
-            eval { $package = $self->pjson($component)->{name} };
-            if ($@) {
-                warning "Unable to load $component: $@";
-                next;
-            }
-            $packages{$component} = $package;
-        }
-        else {
-            warning "Can't find $component directory in " . `pwd`;
-            next;
-        }
-    }
-    return \%packages;
-}
-
 sub pattern {
     my ( $self, $line ) = @_;
     my ( $p, $pattern );
@@ -162,6 +88,8 @@ sub pattern {
     @dirs = map {
         my $s = quotemeta($_);
         $s =~ s#\\\*#[^/]*#g;
+        $s =~ s#\\\((.*?)\\\)#(?:$1)?#g;
+        $s =~ s#\\\[(.*?)\\\]#[$1]#g;
         while ( $s =~ s/\\\{(.*?)\\\}/_____/ ) {
             my $tmp = '(?:' . join( '|', split( /\\?,/, $1 ) ) . ')';
             $s =~ s/_____/$tmp/;
@@ -212,7 +140,7 @@ sub build_module {
     else {
         print "No build command found, searching known files\n";
         foreach my $i (@knwonBuildFiles) {
-            if ( -e $i->[0] ) {
+            if ( -e "$dir/$i->[0]" ) {
                 push @commands, $i->[1];
             }
         }
@@ -368,7 +296,8 @@ sub install_module {
         }
         $noFilesField = 0;
         push @files,
-          grep { $_ ne "$dir/package.json" } $self->readFilesField(
+          grep { $_ ne "$dir/package.json" and $_ ne "$dir/package.yaml" }
+          $self->readFilesField(
             $dir,
             ( map { s/^/!/; s/^\!\!//; $_ } ( $self->open_file($npmignore) ) )
           );
@@ -377,10 +306,12 @@ sub install_module {
         print qq{No "files" field in $dir/package.json, install all files\n};
         @files = ($dir);
         push @files,
-          grep { $_ ne "$dir/package.json" } $self->readFilesField(
+          grep { $_ ne "$dir/package.json" and $_ ne "$dir/package.yaml" }
+          $self->readFilesField(
             $dir,
             ( map { s/^/!/; s/^\!\!//; $_ } ( $self->open_file($npmignore) ) )
           );
+        print "Files to install: " . join( ', ', @files ) . "\n";
     }
     my @dest;
     my $mainFile =
@@ -396,9 +327,12 @@ sub install_module {
     foreach my $p (@files) {
         my $pattern;
         my $not = 0;
-        if ( $p =~ s/\!// ) {
+        if ( $p =~ s/^\!// ) {
             $not = 1;
         }
+        if ( $p =~ s/\!(.*)$// ) {
+            push @files, "!$p$1";
+        }
         if ( $p =~ m/[\*|\{]/ ) {
             ( $p, $pattern ) = $self->pattern($p);
         }
@@ -423,21 +357,23 @@ sub install_module {
                 #                      - package.json contains a "files" field
                 my @tests = (
 
-                    # Ignore directories
+                    # test0: Ignore directories
                     ( -d $_ ),
 
-                    # Don't parse components sub dir for main module
+                    # test1: Don't parse components sub dir for main module
                     ( $d and $d =~ $re and $skipFiles ),
 
-                    # Ignore our debian/ dir
+                    # test2: Ignore our debian/ dir
                     (
                               $dir eq '.'
                           and $File::Find::dir =~ m{^(?:./)?debian}
                           and $skipFiles
                     ),
 
-                    # Ignore license, changelog, readme,... files
+                    # test3: Ignore license, changelog, readme,... files
                     ( $_ =~ $rootFilesIgnored and $noFilesField and !$d ),
+
+                    # test4:
                     (
                               $_ =~ $rootFilesIgnored
                           and $skipFiles
@@ -445,14 +381,16 @@ sub install_module {
                           and !$d
                     ),
 
-          # Ignore Makefile, .eslint anywhere unless DM specified it explicitely
-                    ( $_ =~ $filesIgnored and ( $pattern or $p !~ /$_$/ ) ),
+                    # test5: Ignore Makefile, .eslint anywhere unless DM
+                    # specified it explicitely
                     ( $_ =~ $filesIgnored and ( $pattern or $p !~ /$_$/ ) ),
 
-                  # Ignore test directory unless specified in package.json#files
-                    ( $d and $d =~ $dirIgnored and $noFilesField ),
+                    # test6: Ignore test directory unless DM specified it
+                    # explicitely
+                    ( $d and $d =~ $dirIgnored and $skipFiles ),
 
-                    # Ignore test.js file unless specified in package.json#files
+                    # test7: Ignore test.js file unless specified in
+                    # package.json#files
                     (
                               $File::Find::dir eq $dir
                           and /^(?:example|test)s?\.js$/i
@@ -469,7 +407,8 @@ sub install_module {
                         push @dest, [ $d, $File::Find::name ];
                         $foundMain    = 1 if $File::Find::name eq $mainFile;
                         $foundPkgJson = 1
-                          if $File::Find::name eq 'package.json';
+                          if $File::Find::name eq 'package.json'
+                          or $File::Find::name eq 'package.yaml';
                     }
 
                     # Debug
@@ -511,16 +450,30 @@ EOF
         }
     }
     unless ($foundPkgJson) {
-        push @dest, [ '', "$dir/package.json" ];
+        push @dest,
+          [
+            '',
+            -e "$dir/package.yaml" ? "$dir/package.yaml" : "$dir/package.json"
+          ];
     }
+
+    # Install files in build dir
     foreach (@dest) {
         $self->doit_in_builddir( 'mkdir', '-p', "$path/$_->[0]" )
           unless doit_noerror( { chdir => $self->get_buildpath },
             'test', '-e', "$path/$_->[0]" );
-        my $mode = ( -x $_->[1] ? '755' : '644' );
-        $self->doit_in_builddir( 'cp', '--reflink=auto', '-a', $_->[1],
-            "$path/$_->[0]/" );
+        my $mode =
+          ( ( -x $_->[1] and $_->[1] !~ $noexecFiles ) ? '755' : '644' );
+        if ( -l $_->[1] ) {
+            $self->doit_in_builddir( 'cp', '--reflink=auto', '-a', $_->[1],
+                "$path/$_->[0]/" );
+        }
+        else {
+            $self->doit_in_builddir( 'install', '-m', $mode, $_->[1],
+                "$path/$_->[0]/" );
+        }
     }
+    buildPackageLock( $dir, "$path/pkgjs-lock.json", 1 );
 }
 
 sub filesFieldToList {
@@ -547,11 +500,14 @@ sub readFilesField {
         s#^\.?/##;
         warning "$dir/package.json#files: $_ does not exists\n"
           unless $nowarn
-          or ( $_ and -e "$dir/$_" or /[\{\*]/ );
+          or ( $_ and -e "$dir/$_" )
+          or /[\{\*\!]/;
         $_ ? "$prefix$dir/$_" : ();
     } @lines;
-    push @files, "$dir/package.json"
-      unless grep { $_ eq "$dir/package.json" } @files;
+    push @files,
+      ( -e "$dir/package.yaml" ? "$dir/package.yaml" : "$dir/package.json" )
+      unless grep { $_ eq "$dir/package.json" or $_ eq "$dir/package.yaml" }
+      @files;
     return @files;
 }
 
@@ -566,36 +522,25 @@ sub toJs {
 
 sub cmp_ordered_list {
     my ( $self, $components ) = @_;
+    my $g = Graph->new;
     $components //= $self->component_list;
-    my @dirs = ( sort keys %$components );
-    my @res;
+    $g->add_vertices( keys %$components );
     if ( -e 'debian/nodejs/build_order' ) {
-        @res = map {
-            my $cmp = $_;
-            error "Unknown component $_" unless $components->{$cmp};
-            @dirs = grep { $_ ne $cmp } @dirs;
-            $cmp;
-        } $self->open_file('debian/nodejs/build_order');
+        $g->add_path( $self->open_file('debian/nodejs/build_order') );
     }
     if ( -e 'debian/nodejs/component_links' ) {
-        my %lt;
-        my %firsts;
         map {
             error "Malformed line $. in debian/nodejs/component_links: $_"
               unless /^([\w\-\.\/]+)\s+([\w\-\.\/]+)$/;
-            $lt{$1}     = $2;
-            $firsts{$1} = 1;
+            $g->add_edge( $1, $2 );
         } $self->open_file('debian/nodejs/component_links');
-        @dirs =
-          sort {
-                ( $lt{$a} and $lt{$a} eq $b ) ? -1
-              : ( $lt{$b} and $lt{$b} eq $a ) ? 1
-              : $firsts{$a}                   ? -1
-              : $firsts{$b}                   ? 1
-              : $a cmp $b
-          } @dirs;
     }
-    push @res, @dirs;
+    my @res;
+    eval { @res = $g->toposort };
+    if ($@) {
+        die
+'Unable to resolve components build order: probably cyclic dependencies';
+    }
     return @res;
 }
 
@@ -609,7 +554,7 @@ sub new {
 
 sub check_auto_buildable {
     my $self = shift;
-    if ( -e 'package.json' ) {
+    if ( -e 'package.json' or -e 'package.yaml' ) {
         return 1;
     }
     return 0;
@@ -617,157 +562,9 @@ sub check_auto_buildable {
 
 # auto_configure step: if component are found, create node_modules links
 sub configure {
-    my $self       = shift;
-    my $components = $self->component_list;
-    my @dirs       = ( sort keys %$components );
-    return
-         unless @dirs
-      or -e 'debian/nodejs/' . $self->main_package . '/extlinks'
-      or -e 'debian/nodejs/' . $self->main_package . '/extcopies'
-      or -d 'debian/build_modules' && $self->main_package eq '.';
-    $self->doit_in_builddir( 'mkdir', 'node_modules' )
-      unless -e 'node_modules';
-
-    # External links
-    foreach my $dir (
-        (
-            $self->main_package
-            ? (
-                $self->main_package ne '.'
-                ? ( $self->main_package, '.' )
-                : ('.')
-            )
-            : ('.')
-        ),
-        @dirs
-      )
-    {
-        if ( -e "debian/nodejs/$dir/extlinks" ) {
-            map {
-                my ( $in, $src );
-                $_ =~ s/\s+(\S*).*?$//;
-                if (
-                        $1
-                    and $1 eq 'test'
-                    and (
-                        (
-                                $ENV{DEB_BUILD_OPTIONS}
-                            and $ENV{DEB_BUILD_OPTIONS} =~ /\bnocheck\b/
-                        )
-                        or (    $ENV{DEB_BUILD_PROFILES}
-                            and $ENV{DEB_BUILD_PROFILES} =~ /\bnocheck\b/ )
-                    )
-                  )
-                {
-                    print STDERR "Skipping $_ link\n";
-                }
-                else {
-                    run [ 'nodepath', $_ ], \$in, \$src;
-                    if ( $? or not $src ) {
-                        print STDERR
-                          "### $_ is required by debian/nodejs/$dir/extlinks"
-                          . " but not available\n";
-                        exit 1;
-                    }
-                    else {
-                        chomp $src;
-                        my $tmp = $src;
-                        $tmp =~ s#^.*?nodejs/##;
-                        $tmp =~ s#/.*?$## or $tmp = '';
-                        $tmp = "$tmp/" if $tmp;
-                        $self->dolink( $src, "$dir/node_modules/$tmp" );
-                    }
-                }
-            } $self->open_file("debian/nodejs/$dir/extlinks");
-        }
-        if ( -e "debian/nodejs/$dir/extcopies" ) {
-            map {
-                my ( $in, $src );
-                $_ =~ s/\s+(\S*).*?$//;
-                if (
-                        $1
-                    and $1 eq 'test'
-                    and (
-                        (
-                                $ENV{DEB_BUILD_OPTIONS}
-                            and $ENV{DEB_BUILD_OPTIONS} =~ /\bnocheck\b/
-                        )
-                        or (    $ENV{DEB_BUILD_PROFILES}
-                            and $ENV{DEB_BUILD_PROFILES} =~ /\bnocheck\b/ )
-                    )
-                  )
-                {
-                    print STDERR "Skipping $_ copy\n";
-                }
-                else {
-                    run [ 'nodepath', $_ ], \$in, \$src;
-                    if ( $? or not $src ) {
-                        print STDERR
-                          "### $_ is required by debian/nodejs/$dir/extcopies"
-                          . " but not available\n";
-                        exit 1;
-                    }
-                    else {
-                        chomp $src;
-                        my $dst = $src;
-                        $dst =~ s#.*?nodejs/##;
-                        if ( $dst =~ s#/[^/]*$## ) {
-                            $self->doit_in_builddir( 'mkdir', '-p',
-                                "$dir/node_modules/$dst" );
-                        }
-                        else {
-                            $dst = '';
-                        }
-                        $self->doit_in_builddir( 'cp', '-rL', $src,
-                            "$dir/node_modules/$dst" );
-                    }
-                }
-            } $self->open_file("debian/nodejs/$dir/extcopies");
-        }
-    }
-
-    # Link each component in node_modules
-    foreach my $component (@dirs) {
-        my $package = $components->{$component};
-        $self->dolink( "$component", "node_modules/$package" )
-          unless -e "debian/nodejs/$component/nolink";
-    }
-
-    # Links between components
-    if ( -e 'debian/nodejs/component_links' ) {
-        my %links = map {
-            error "Malformed line $. in debian/nodejs/component_links: $_"
-              unless /^([\w\-\.\/]+)\s+([\w\-\.\/]+)$/;
-            my ( $src, $dest ) = ( $1, $2 );
-            foreach my $s ( $src, $dest ) {
-                error "component_links: Unknown component $s"
-                  unless $components->{$s};
-            }
-            ( $src => $dest );
-        } $self->open_file('debian/nodejs/component_links');
-        foreach ( keys %links ) {
-            $self->dolink( $_, "$links{$_}/node_modules/$components->{$_}" );
-        }
-    }
-
-    # Links all modules found in debian/build_modules
-    if ( -d 'debian/build_modules' ) {
-        opendir my $buildDir, 'debian/build_modules';
-        my @buildMods = grep { /^[^\.]/ } ( readdir $buildDir );
-        foreach my $mod (@buildMods) {
-            if ( $mod =~ /^\@/ ) {
-                opendir my $buildDir, "debian/build_modules/$mod";
-                my @subMods = grep { /^[^\.]/ } ( readdir $buildDir );
-                $self->dolink( "debian/build_modules/$mod/$_",
-                    "node_modules/$mod/$_" )
-                  foreach (@subMods);
-            }
-            else {
-                $self->dolink( "debian/build_modules/$mod",
-                    "node_modules/$mod" );
-            }
-        }
-    }
+    link_external_modules();
+    link_internal_modules();
+    link_build_modules();
 }
 
 # build step
@@ -783,7 +580,7 @@ sub build {
         warning "Aborting auto_build\n";
 
         # To avoid hard transition, this step never fail
-        #exit 1;
+        exit 1;
     }
 }
 
@@ -793,24 +590,10 @@ sub test {
     my $components = $self->component_list;
     my @dirs       = ( sort keys %$components );
     my @testlinks;
-    if ( -d 'debian/tests/test_modules' ) {
-        opendir my $testDir, 'debian/tests/test_modules';
-        my @testMods = grep { /^[^\.]/ } ( readdir $testDir );
-        foreach my $mod (@testMods) {
-            if ( $mod =~ /^\@/ ) {
-                opendir my $testDir, "debian/tests/test_modules/$mod";
-                my @subMods = grep { /^[^\.]/ } ( readdir $testDir );
-                $self->dolink( "debian/tests/test_modules/$mod/$_",
-                    "node_modules/$mod/$_" )
-                  foreach (@subMods);
-                push @testlinks, map { "$mod/$_" } @subMods;
-            }
-            else {
-                $self->dolink( "debian/tests/test_modules/$mod",
-                    "node_modules/$mod" );
-                push @testlinks, $mod;
-            }
-        }
+    my @testMods = list_test_modules();
+    foreach my $mod (@testMods) {
+        $self->dolink( "debian/tests/test_modules/$mod", "node_modules/$mod" );
+        push @testlinks, $mod;
     }
     eval {
         if ( my $p = $self->main_package ) {
@@ -834,10 +617,17 @@ sub test {
     if ( -e $testfile ) {
         $self->doit_in_builddir( '/bin/sh', '-ex', $testfile );
     }
-    else {
-        $self->doit_in_builddir( '/usr/bin/node', '-e',
-            'require("./' . $self->main_package . '")' )
-          if $self->main_package;
+    elsif ( $self->main_package ) {
+        if (    $self->pjson( $self->main_package )->{type}
+            and $self->pjson( $self->main_package )->{type} eq 'module' )
+        {
+            print
+"Found type=module, skipping require test (will do in autopkgtest)\n";
+        }
+        else {
+            $self->doit_in_builddir( '/usr/bin/node', '-e',
+                'require("./' . $self->main_package . '")' );
+        }
     }
     foreach (@testlinks) {
         print "Removing node_modules/$_\n";
@@ -852,7 +642,7 @@ sub test {
 sub install {
     my $self    = shift;
     my $destdir = shift;
-    my @pkgs    = getpackages();
+    my @pkgs    = grep { /\w/ } getpackages();
     my ($package) =
       $#pkgs ? ( grep { $_ =~ /^node-/ } getpackages() ) : @pkgs;
     ($package) = @pkgs unless ($package);
@@ -900,7 +690,7 @@ EOT
             %root_cmp = map {
                 my $s = $_;
                 map { ( $_ => 1 ) } glob($s)
-            } $self->open_file('debian/nodejs/root_modules');
+            } @lines;
         }
     }
     elsif ( !$self->main_package ) {
@@ -949,9 +739,7 @@ EOT
             $dir
         );
         if ( $root_cmp{$dir} ) {
-            my $tmp = 'node-' . $self->pjson($dir)->{name};
-            $tmp =~ s/[\@_\/]/-/g;
-            $tmp =~ s/--*/-/g;
+            my $tmp = 'node-' . normalize_name( $self->pjson($dir)->{name} );
             my $v;
             if ( defined( $v = $self->pjson($dir)->{version} ) ) {
                 $tmp .= " (= $v)";
@@ -965,8 +753,13 @@ EOT
     if ( $package ne $pname ) {
         my $tmp = $pname;
         my $v;
-        if ( -e $self->main_package . "/package.json"
-            and defined( $v = $self->pjson( $self->main_package )->{version} ) )
+        if (
+            (
+                   -e $self->main_package . "/package.json"
+                or -e $self->main_package . "/package.yaml"
+            )
+            and defined( $v = $self->pjson( $self->main_package )->{version} )
+          )
         {
             $tmp .= " (= $v)";
         }
@@ -978,39 +771,46 @@ EOT
     if (@provides) {
         print "Populate \${nodejs:Provides}:\n";
         print " + $_\n" foreach (@provides);
-        addsubstvar( $package, 'nodejs:Provides', join( ', ', @provides ) );
+        foreach my $pkg (@pkgs) {
+            addsubstvar( $pkg, 'nodejs:Provides', join( ', ', @provides ) );
+        }
+    }
+    if ( my $_builtUsing = builtUsing() ) {
+        foreach my $pkg (@pkgs) {
+            addsubstvar( $pkg, 'nodejs:BuiltUsing', $_builtUsing );
+        }
     }
 
-    {
-        my ( $in, $src );
-        if ( -e 'debian/node-node-expat.substvars' ) {
-            run [
-                'grep', '-Po',
-                'shlibs:Depends=.*libnode\d+ \(>= \K\S+(?=\))',
-                'debian/node-node-expat.substvars'
-              ],
-              \$in, \$src;
-        }
-        if ( not $src ) {
-            run [
-                'dpkg-query', '--showformat=${source:Upstream-Version}',
-                '--show',     'nodejs'
-              ],
-              \$in, \$src;
-        }
-        if ( not $src or $? ) {
-            print STDERR "### node version not found\n";
-        }
-        else {
-            chomp $src;
-            addsubstvar( $package, 'nodejs:Version', $src );
-            print "Set \${nodejs:Version} to $src\n";
+    my ( $in, $src );
+    if ( -e 'debian/node-node-expat.substvars' ) {
+        run [
+            'grep', '-Po',
+            'shlibs:Depends=.*libnode\d+ \(>= \K\S+(?=\))',
+            'debian/node-node-expat.substvars'
+          ],
+          \$in, \$src;
+    }
+    if ( not $src ) {
+        run [
+            'dpkg-query', '--showformat=${source:Upstream-Version}',
+            '--show',     'nodejs'
+          ],
+          \$in, \$src;
+    }
+    if ( not $src or $? ) {
+        print STDERR "### node version not found\n";
+    }
+    else {
+        chomp $src;
+        foreach my $pkg (@pkgs) {
+            addsubstvar( $pkg, 'nodejs:Version', $src );
         }
+        print "Set \${nodejs:Version} to $src\n";
     }
 
     # Links
     if ( -e "debian/nodejs/links" ) {
-        my $tmp = tmpdir($package);
+        my $tmp = $self->auto_install_dir($package);
         map {
             my $line = $_;
             my ( $src, $dest ) =
@@ -1030,47 +830,34 @@ sub clean {
     my $self       = shift;
     my $components = $self->component_list;
     my @dirs       = ( sort keys %$components );
-    foreach my $dir ( $self->main_package, @dirs ) {
+    foreach my $dir ( main_package, @dirs ) {
         $self->doit_in_sourcedir( 'rm', '-rf', "$dir/node_modules/.cache" );
-        if ( -e "debian/nodejs/$dir/extlinks" ) {
-            map { $self->do_unlink("$dir/node_modules/$_"); }
-              $self->open_file("debian/nodejs/$dir/extlinks");
-        }
-        if ( -e "debian/nodejs/$dir/extcopies" ) {
-            map {
-                $self->doit_in_builddir( 'rm', '-rf', "$dir/node_modules/$_" );
-            } $self->open_file("debian/nodejs/$dir/extcopies");
-        }
     }
-    if ( -e 'debian/nodejs/component_links' ) {
-        map {
-            error "Malformed line $. in debian/nodejs/component_links: $_"
-              unless /^([\w\-\.\/]+)\s+([\w\-\.\/]+)$/;
-            my ( $src, $dest ) = ( $1, $2 );
-            $self->do_unlink("$dest/node_modules/$components->{$src}");
-        } $self->open_file('debian/nodejs/component_links');
-    }
-    foreach my $component ( $self->main_package, @dirs ) {
-        my $package = $components->{$component};
+    clean_external_modules;
+    clean_internal_modules;
+    clean_build_modules();
+    clean_test_modules();
 
-        # Error not catched here
-        $self->do_unlink("node_modules/$package") if $package;
+    # Clean test link
+    if ( my $p = $self->main_package ) {
+        $self->do_unlink( 'node_modules/' . $self->pjson($p)->{name} );
     }
-    if ( -d 'debian/build_modules' ) {
-        opendir my $buildDir, 'debian/build_modules';
-        my @buildMods = grep { /^[^\.]/ } ( readdir $buildDir );
-        foreach my $mod (@buildMods) {
-            if ( $mod =~ /^\@/ ) {
-                opendir my $buildDir, "debian/build_modules/$mod";
-                my @subMods = grep { /^[^\.]/ } ( readdir $buildDir );
-                $self->do_unlink("node_modules/$mod/$_") foreach (@subMods);
-                eval { rmdir "node_modules/$mod" };
-            }
-            else {
-                $self->do_unlink("node_modules/$mod");
-            }
-        }
+
+    # Try to clean node_modules
+    rmdir 'node_modules';
+}
+
+sub auto_install_dir {
+    my ( $self, $package ) = @_;
+    my $destdir;
+    my @allpackages = getpackages();
+    if ( @allpackages > 1 or not compat(14) ) {
+        $destdir = "debian/tmp";
     }
+    else {
+        $destdir = tmpdir($package);
+    }
+    return $destdir;
 }
 
 1;
